PIPELINE_RUN_ID = "pipeline_run_id"
root_span_suffix = "_root_span"
STARTED = "Started"
ENDED = "Ended"
DURATION = "Duration"
TRIGGER_RULE = "Trigger Rule"
PIPELINE_NAME_SUFFIX = "_pipeline"
PIPELINE_STATE = "pipeline_state"
ROOT_SPAN_SUFFIX = "_root_span"
STATE = "state"
FAILED = "failed"
SUCCESS = "success"
MESSAGE = "message"
DAG_ID = "airflow_dag_id"
TASK_ID = "airflow_task_id"
DAG_RUN_ID = "airflow_dag_run_id"
TASK_RUN_ID = "airflow_task_run_id"
OPERATOR = "operator"
TRY_NUMBER = "try_number"
MAP_INDEX = "map_index"
DURATION_UNIT = " seconds"
RUN_TYPE = "run_type"
SCHEDULE_INTERVAL = "schedule_interval"
OWNER = "owner"
QUEUE = "queue"
DAG_URL = "url"
INVALID_ENV_SELECTION_MESSAGE = "Environment variables --OBSERVE DAG and --IGNORE DAG are mutually exclusive.  " \
                                " Please remove one option to observe/ignore the DAG. "
